/*
 *  double.cpp
 *  This is a simple double buffered program.
 *  Pressing the left mouse button rotates the square.
 *  Pressing the middle or right mouse button stops the rotation.
 */


#include "Angel.h"


/* spin is the square's rotation angle.  It's updated by spinDisplay(),
 * which is registered as the idle callback by the mouse callback with
 * a left mouse click.  A middle or right mouse click de-registers the
 * idle callback.  Look at mouse() to see how this is accomplished.
 * spin's units are degress.
 */

GLfloat spin = 0.0;


/* The amount, in degrees, to increase spin by each time spinDisplay()
 * is called.
 */

GLfloat spinDelta = 0.5;


/* Handles for the projection and model view uniform variables in the
 * vertex shader.
 */

GLuint projection;
GLuint model_view;


void display(void)
{
   mat4 mv;   /* Model view matrix */

   glClear(GL_COLOR_BUFFER_BIT);

   /* RotateZ() constructs a rotation matrix using the Z axis as the
    * axis of rotation.  Note that the units of the rotation angle are
    * degrees, not radians.
    */

   mv = RotateZ(spin);
   glUniformMatrix4fv(model_view, 1, GL_TRUE, mv);

   glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

   glutSwapBuffers();
}


void spinDisplay(void)
{
   spin = spin + spinDelta;
   if (spin > 360.0)
      spin = spin - 360.0;

   glutPostRedisplay();
}


void init(void) 
{
   /* The square's color.  It will be passed to the vertex shader as a
    * uniform variable.  Once passed to the vertex shader, it is no
    * longer needed, so we only need a local variable.
    */

   vec4 color = vec4(1.0, 1.0, 1.0, 1.0);

   /* The square's vertices.  This data will be buffered in the GL
    * server, so we only need a local variable for this data.
    */

   vec2 points[4] = { vec2(-25.0, -25.0), vec2(25.0, -25.0),
                      vec2(25.0, 25.0), vec2(-25.0, 25.0)
                    };

   glClearColor (0.0, 0.0, 0.0, 1.0);
   glShadeModel (GL_FLAT);

   /* Compile and link the shader programs. */

   GLuint program = InitShader("vshader.glsl", "fshader.glsl");

   /* Bind the shader programs. */
   
   glUseProgram(program);

   /* Get the locations of the uniform variables in the vertex shader.
    *
    * IMPORTANT NOTE: Observe that color is declared as a vec4 above,
    * vColor is declared in the vertex shader as a vec4, and that color
    * is assigned to vColor by a call to glUniform4fv().  The source,
    * color, is a 4-tuple, the target, vColor, is a 4-tuple, and we're
    * copying a 4-tuple (the "4fv" in glUniform4fv()) from source to
    * target.  Missing any of these details will likely cause color
    * data to not be drawn.
    */

   GLuint vColor = glGetUniformLocation(program, "vColor");
   glUniform4fv(vColor, 1, color);

   /* The model_view uniform variable receives its value in display(). */

   model_view = glGetUniformLocation(program, "model_view");

   /* The projection uniform variable receives its value in reshape(). */

   projection = glGetUniformLocation(program, "projection");

   /* This VAO remains bound throughout the execution of the program,
    * hence we only need a local variable to perform the one and only
    * VAO binding.
    */

   GLuint vao;
   glGenVertexArrays(1, &vao);
   glBindVertexArray(vao);

   GLuint buffer;
   glGenBuffers(1, &buffer);
   glBindBuffer(GL_ARRAY_BUFFER, buffer);
   glBufferData(GL_ARRAY_BUFFER, sizeof(points), points, GL_STATIC_DRAW);

   GLuint loc = glGetAttribLocation(program, "vPosition");
   glEnableVertexAttribArray(loc);
   glVertexAttribPointer(loc, 2, GL_FLOAT, GL_FALSE, 0,
                         BUFFER_OFFSET(0));
}


void reshape(int w, int h)
{
   mat4 p;    /* Projection matrix */

   glViewport (0, 0, (GLsizei) w, (GLsizei) h);

   /* It strikes me that this is overkill, since we aren't doing anything
    * to correct aspect ratio mismatches.
    */

   p = Ortho(-50.0, 50.0, -50.0, 50.0, -1.0, 1.0);
   glUniformMatrix4fv(projection, 1, GL_TRUE, p);
}


void mouse(int button, int state, int x, int y) 
{
   switch (button) {

      case GLUT_LEFT_BUTTON:
         if (state == GLUT_DOWN)
            glutIdleFunc(spinDisplay);

         break;

      case GLUT_MIDDLE_BUTTON:
      case GLUT_RIGHT_BUTTON:
         if (state == GLUT_DOWN)
            glutIdleFunc(NULL);

         break;

      default:
         break;
   }
}

   
/* 
 *  Request double buffer display mode.
 *  Register mouse input callback functions
 */

int main(int argc, char** argv)
{
   glutInit(&argc, argv);
   glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB);
   glutInitWindowSize (250, 250); 
   glutInitWindowPosition (100, 100);
   glutInitContextVersion(3, 2);
   glutInitContextProfile(GLUT_CORE_PROFILE);
   glutCreateWindow ("Double");

   glewExperimental = GL_TRUE;
   glewInit();

   init ();

   glutDisplayFunc(display); 
   glutReshapeFunc(reshape); 
   glutMouseFunc(mouse);

   glutMainLoop();

   return 0;
}
